"use strict";
const fs_extra_p_1 = require("fs-extra-p");
const path = require("path");
const bluebird_lst_c_1 = require("bluebird-lst-c");
const ts = require("typescript");
const __awaiter = require("./awaiter");
const filenameToMid = (function () {
    if (path.sep === '/') {
        return (it) => it;
    }
    else {
        const separatorExpression = new RegExp(path.sep.replace('\\', '\\\\'), 'g');
        return (it) => it.replace(separatorExpression, '/');
    }
})();
function generateDeclarationFile(moduleName, declarationFiles, compilerOptions, out, basePath) {
    return __awaiter(this, void 0, void 0, function* () {
        const fileNameToModuleId = {};
        const relativeOutDir = path.relative(basePath, compilerOptions.outDir);
        yield fs_extra_p_1.mkdirs(path.dirname(out));
        const eol = "\n";
        const indent = "  ";
        const output = fs_extra_p_1.createWriteStream(out, { mode: parseInt('644', 8) });
        return new bluebird_lst_c_1.default((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            output.on('close', resolve);
            output.on('error', reject);
            for (let sourceFile of declarationFiles) {
                writeDeclaration(sourceFile, compilerOptions, relativeOutDir);
            }
            output.end();
        }));
        function writeDeclaration(declarationFile, compilerOptions, relativeOutDir) {
            if (declarationFile.text.length === 0) {
                return;
            }
            let sourceModuleId;
            let baseName = "";
            const fileNameWithoutExt = declarationFile.fileName.slice(0, -5).replace(/\\/g, "/");
            const name = fileNameWithoutExt.substring(compilerOptions.outDir.length + 1);
            if (moduleName) {
                baseName = moduleName + '/';
                sourceModuleId = moduleName;
                if (name !== "index") {
                    sourceModuleId += '/' + relativeOutDir;
                }
            }
            else {
                sourceModuleId = relativeOutDir;
            }
            baseName += relativeOutDir;
            const parentDir = name.includes("/") ? `${baseName}/${path.dirname(name)}` : baseName;
            if (name !== "index") {
                sourceModuleId += '/' + name;
            }
            if (declarationFile.fileName.endsWith("main.d.ts")) {
                sourceModuleId = moduleName;
            }
            output.write(`declare module "${sourceModuleId}" {${eol}${indent}`);
            fileNameToModuleId[path.resolve(fileNameWithoutExt).replace(/\\/g, "/")] = sourceModuleId;
            const content = processTree(declarationFile, (node) => {
                if (node.kind === ts.SyntaxKind.ExternalModuleReference) {
                    const expression = node.expression;
                    if (expression.text.charAt(0) === '.') {
                        return ' require(\'' + filenameToMid(path.join(path.dirname(sourceModuleId), expression.text)) + '\')';
                    }
                }
                else if (node.kind === ts.SyntaxKind.DeclareKeyword) {
                    return '';
                }
                else if (node.kind === ts.SyntaxKind.StringLiteral && (node.parent.kind === ts.SyntaxKind.ExportDeclaration || node.parent.kind === ts.SyntaxKind.ImportDeclaration)) {
                    const text = node.text;
                    if (text.charAt(0) === '.') {
                        if (text.charAt(1) === '.') {
                            const m = fileNameToModuleId[path.resolve(path.dirname(declarationFile.fileName), text).replace(/\\/g, "/")];
                            if (m != null) {
                                return ` "${m}"`;
                            }
                            return ` "${baseName}/${text.substring(3)}"`;
                        }
                        else {
                            return ` "${parentDir}/${text.substring(2)}"`;
                        }
                    }
                }
                return null;
            });
            let prev = content.replace(new RegExp(eol + '(?!' + eol + '|$)', 'g'), '$&' + indent);
            prev = prev.replace(/;/g, '');
            if (indent != "    ") {
                prev = prev.replace(/    /g, indent);
            }
            output.write(prev);
            if (prev.charAt(prev.length - 1) != '\n') {
                output.write(eol);
            }
            output.write('}' + eol + eol);
        }
    });
}
exports.generateDeclarationFile = generateDeclarationFile;
function processTree(sourceFile, replacer) {
    let code = '';
    let cursorPosition = 0;
    function skip(node) {
        cursorPosition = node.end;
    }
    function readThrough(node) {
        code += sourceFile.text.slice(cursorPosition, node.pos);
        cursorPosition = node.pos;
    }
    function visit(node) {
        readThrough(node);
        if (node.flags & ts.ModifierFlags.Private) {
            skip(node);
            return;
        }
        if (node.kind === ts.SyntaxKind.ImportDeclaration && node.importClause == null) {
            skip(node);
            return;
        }
        const replacement = replacer(node);
        if (replacement != null) {
            code += replacement;
            skip(node);
        }
        else {
            if (node.kind === ts.SyntaxKind.ClassDeclaration || node.kind === ts.SyntaxKind.InterfaceDeclaration || node.kind === ts.SyntaxKind.FunctionDeclaration) {
                code += "\n";
            }
            ts.forEachChild(node, visit);
        }
    }
    visit(sourceFile);
    code += sourceFile.text.slice(cursorPosition);
    return code;
}
//# sourceMappingURL=declarationGenerator.js.map